#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include <malloc.h>
#include "misc.h"
#include "file.h"
#include "xy.h"
#include "nsc.h"
#include "genitem.h"
#include "edb.h"
#include "edbproc.h"	/* public  structures */
#include "edbpriv.h"	/* private structures */

#ifdef Debug
#define HelpMe
#define DebugPrint(x) printf x;
#define DIAG(x)	x
#else
#define DebugPrint(x)
#define DIAG(x)
#endif

#define MAXNAMELEN 80

#define	SF_ALLFIELDS		0
#define SF_KEYFIELDSONLY	1


typedef union {
		s_char **ccp;
		s_char *cp;
		u_char *up;
		int *ip;
		short *sp;
		unsigned short *usp;
		long *lp;
		float *fp;
		time_t *tp;
	} Thing;

extern ObjectDescriptor objTable[];

/*
	findObject -- find object type in our object table
*/
ObjectDescriptor *findObject(code)
	int code;
{
	ObjectDescriptor *p;
	for (p=objTable; p->objtype!=EF_BAD; p++)
		if (p->objtype==code)
			return p;
	return 0;
}


/*
	 newField -- allocate a new Field structure
 */
Field *newField(name, kind, count)
	s_char *name;
	int kind;
	int count;
{
	Field *f;
	s_char *fname;
	int namelen;
	fname = (s_char *) malloc(strlen(name)+1);
	assert(fname!=0);
	strcpy(	fname, name);
	f = (Field *) malloc(sizeof(Field));
	assert(f!=0);
	f->fldname = fname;
	f->fldtype = kind;
	f->fldcount = count;
	f->next = 0;
	return f;
}

/*
	copyField -- copy constructor for Field structure
*/
Field *copyField(aField)
	Field *aField;
{
	return newField(aField->fldname, aField->fldtype, aField->fldcount);
}

/*
	newDict -- allocate a new Dictionary structure
*/
Dictionary *newDict(num)
	int num;
{
	Dictionary *d;
	d = (Dictionary *) malloc(sizeof(Dictionary));
	assert(d!=0);
	d->next = 0;
	d->objnum = num;
	d->flds = 0;
	return d;
}

/*
	findDict -- find a Dictionary entry in the database
*/
Dictionary *findDict(edb, code)
	Empire_Database *edb;
	int code;
{
	Dictionary *p;
	for (p = (Dictionary *)edb->dictionary ; p ; p=p->next) {
		if (p->objnum==code) return p;
	}
	return 0;
}
	
/*
	insertDict -- insert a Dictionary entry into the database
	(simple push down with no check for duplicates)
*/
void insertDict(edb, p)
	Empire_Database *edb;
	Dictionary *p;
{
	assert(edb!=0);
	if (!p) return;
	p->next = (Dictionary *) edb->dictionary;
	edb->dictionary = p;
}

/*
	insertField -- add a field to an object definition
*/
void insertField(d, f)
	Dictionary *d;
	Field *f;
{
	Field *prev, *p;

	assert(d!=0);

	if (!f) return;
	
	for (prev=0,p=d->flds; p; prev=p,p=p->next)
		/**/ ;
	if (prev)
		prev->next = f;
	else
		d->flds = f;
	f->next = 0;
}

Field *findField(d, name)
	Dictionary *d;
	s_char *name;
{
	Field *p;
	assert(d!=0);
	for (p=d->flds; p; p=p->next)
		if (strcmp(name,p->fldname)==0)
			return p;
	return NULL;
}

/*
	getnum -- parse a number in the C_SYNC message
*/
int getnum(pp)
	s_char **pp;
{
	s_char *p;
	int n;
	p = *pp;
	while (*p==' ') p++;
	n = 0;
	while (isdigit(*p)) {
		n = n*10 + (*p++) - '0';
	}
	*pp = p;
	return n;
}

/*
	parseField -- parse a field entry in an object definition
	allocates a Field structure and return it if successful.
*/

void parseField(dict, msg)
	Dictionary *dict;
	s_char **msg;
{
	/* we're parsing entries that look like this */
	/*   fieldname,num,num:fieldname,num,num */
	char namebuf[MAXNAMELEN];
	Field *fld;
	s_char *p, *nameptr;
	int namelen, fieldkind;
	int fieldcount = 0;
	nameptr = p = *msg;
	/* find end of fieldname */
	while ( *p!=0 && *p!=':' && *p!=',') p++;
	namelen = p - *msg;
	if (namelen>=MAXNAMELEN-5)
		namelen = MAXNAMELEN-5;
	strncpy(namebuf,nameptr,namelen);
	namebuf[namelen] = 0;
	if (*p!=',') return;	/* probably bogus */
	p++;	/* skip comma */
	fieldkind = getnum(&p);
	if (*p==',') {
		p++;
		fieldcount = getnum(&p);
	}
	while ( *p!=0 && *p!=':') p++;	/* skip any other junk */
	*msg = p;
	if (findField(dict, namebuf)) {
		DebugPrint(("csync: warning: duplicate field def for object %d: %s\n",
				dict->objnum, namebuf))
		strcat(namebuf,"_dup");
	}
	fld = newField( namebuf, fieldkind, fieldcount);
	insertField(dict, fld);
}

/*
	findObjectField -- match fields received with fields known
	This is done only when definitions are received, not
	when objects are received.
*/
struct castr *findObjectField(objca, fldname)
	struct castr *objca;
	s_char *fldname;
{
	for (;objca->ca_code>0;objca++)
		if (objca->ca_name==0)
			return 0;
		else if (strcmp(objca->ca_name, fldname)==0)
			return objca;
	return 0;
}

void bindFieldDefinitions(dict)
	Dictionary *dict;
{
	ObjectDescriptor *obj;
	Field *f;
	dict->objdesc = obj = findObject(dict->objnum);
	if (!obj) {
		DebugPrint(("csync: warning: unknown object type %d\n", dict->objnum));
		return;	/* an object we cannot handle */
	}
	for (f=dict->flds;f;f=f->next) {
		f->boundto = findObjectField(obj->obj_ca, f->fldname);
#ifdef SERVERTYPES
		if (!f->boundto)
			if (obj->hasVarfields)
				f->boundto = findObjectField(var_ca, f->fldname);
#endif
		if (!f->boundto) {
			DebugPrint(("csync: warning: field not bound: %s\n", f->fldname));
		}
	}
}


/*
	parse_typedef -- parse a C_SYNC type definition
*/
void parse_typedef(edb, msg)
	Empire_Database *edb;
	s_char *msg;
{
	int objnum, baseobjnum;
	Dictionary *baseobj;
	Dictionary *obj;
	Field *fld;
	ObjectDescriptor *od;

	baseobjnum = getnum(&msg);
	objnum = getnum(&msg);
	if (baseobjnum==objnum)
		baseobj = NULL;
	else {
		baseobj = findDict(edb, baseobjnum);
		DIAG( if (!baseobj) {
			DebugPrint(("csync: typedef, unknown base object %d for object %d\n",
				baseobjnum, objnum));
		} )
	}
	if (findDict(edb, objnum)) return;	/* already done */

	DIAG(
		od = findObject(objnum);

		printf("csync: defining object type: %d (%s)\n",
			objnum, od?od->objdescr:"unknown");
	)
	
	obj = newDict(objnum);
	
#if (0)
	/* base object fields are included with the definition for now */
	if (baseobj) {
		for (fld=baseobj->flds;fld;fld=fld->next) {
			insertField(obj, copyField(fld));
		}
	}
#endif
	/* now parse fields and field types */
	while (*msg==' ') msg++;
	for (;;) {
		 parseField(obj, &msg);
		if (*msg != ':') break;
		msg++;
	}
	/* bind fields in the csync message to fields in the database */
	bindFieldDefinitions(obj);

	insertDict(edb, obj);

}

void storeField(obj, item, fld, buf, rectype, subscript)
	ObjectDescriptor *obj;
	struct genobject *item;
	Field *fld;
	s_char *buf;
	int rectype;
	int subscript;
{
	/* msg is a string delimited with a ":" or a newline or a null */
	
	int fldoffs;
	long fldkind;
	
	Thing thing;
	struct castr *captr;
	
	/* get object descriptor */	
	captr = fld->boundto;
	if (!captr) {
		return;
	}

	fldoffs = (captr->ca_code)&0xffff;
	fldkind = (captr->ca_code)&NSC_TMASK;
	if (captr->ca_code&NSC_VAR) {
#ifdef SERVERTYPES
		/* this routine only exists if we're in the server... */
		if (obj && obj->putvar)
			obj->putvar(fldoffs, atoi(buf), item);
#endif
		return;
	}
	if (subscript) {
		if (subscript<0 || subscript>=captr->ca_len) {
			DIAG(
				printf("Subscript out of range: %s[%d] %08lx\n", 
					fld->fldname, subscript, fldkind);
			)
			subscript=0;
		}
	}
	
	thing.cp = ((char *)item) + fldoffs;
	switch (fldkind) {
		case NSC_CHAR:
			if (captr->ca_len==0) 
				*thing.cp = atoi(buf);
			else {
				strncpy(thing.cp,buf,captr->ca_len);
				thing.cp[captr->ca_len-1] = '\0';
			}
			break;
		case NSC_UCHAR:
			thing.up[subscript] = atoi(buf);
			break;
		case NSC_SHORT:
			thing.sp[subscript] = atoi(buf);
			break;
		case NSC_INT:
			thing.ip[subscript] = atoi(buf);
			break;
		case NSC_USHORT:
			thing.usp[subscript] = atoi(buf);
			break;
		case NSC_LONG:
			thing.lp[subscript] = atol(buf);
			break;
		case NSC_XCOORD:
		case NSC_YCOORD:
			thing.sp[subscript] = atoi(buf);
			break;
		case NSC_FLOAT:
			thing.fp[subscript] = atof(buf);
			break;
		case NSC_CHARP:
			if (*thing.ccp && fldoffs>0)
				free(*thing.ccp);
			*thing.ccp = strdup(buf);
			break;
#ifdef NSC_TIME
		case NSC_TIME:
			thing.tp[subscript]= (time_t) atol(buf);
			break;
#endif
		default:
			DIAG(
				printf("unknown field kind: %s %08lx\n", 
					fld->fldname, fldkind);
			)
			break;
	}
/*	printf("%4d: %10s = %-10s\n", fldoffs, fld->fldname, buf); */
}

void sprintField(obj, pbuf, item, fld, rectype, subscript)
	ObjectDescriptor *obj;
	char *pbuf;
	struct genobject *item;
	Field *fld;
	int rectype;
	int subscript;
{
	/* msg is a string delimited with a ":" or a newline or a null */
	
	int fldoffs;
	long fldkind;
	int  val;
	Thing thing;
	struct castr *captr;
	
	/* get object descriptor */	
	captr = fld->boundto;
	if (!captr) {
		return;
	}

	fldoffs = (captr->ca_code)&0xffff;
	fldkind = (captr->ca_code)&NSC_TMASK;
	
	if (captr->ca_code&NSC_VAR) {
#ifdef SERVERTYPES
		/* this routine only exists if we're in the server... */
		if (obj && obj->getvar)
			val = obj->getvar(fldoffs, item);
		else
			val = 0;
		sprintf(pbuf,"%d", val);
#endif
		return;
	}
	if (subscript) {
		if (subscript<0 || subscript>=captr->ca_len) {
			strcpy(pbuf,"*error*");
			return;
		}
	}
	thing.cp = ((char *)item) + fldoffs;
	switch (fldkind) {
		case NSC_CHAR:
			sprintf(pbuf, "%d", thing.cp[subscript]);
			break;
		case NSC_UCHAR:
			sprintf(pbuf, "%d", thing.up[subscript]);
			break;
		case NSC_SHORT:
			sprintf(pbuf, "%d", thing.sp[subscript]);
			break;
		case NSC_INT:
			sprintf(pbuf, "%d", thing.ip[subscript]);
			break;
		case NSC_USHORT:
			sprintf(pbuf, "%d", thing.usp[subscript]);
			break;
		case NSC_LONG:
			sprintf(pbuf, "%ld", thing.lp[subscript]);
			break;
		case NSC_XCOORD:
		case NSC_YCOORD:
			sprintf(pbuf, "%d", thing.sp[subscript]);
			break;
		case NSC_FLOAT:
			sprintf(pbuf, "%f", thing.fp[subscript]);
			break;
		case NSC_CHARP:
			strcpy(pbuf, *thing.ccp);
			break;
#ifdef NSC_TIME
		case NSC_TIME:
			strftime(pbuf, 20, "%s", thing.tp[subscript]);
			break;
#endif
		default:
			strcpy(pbuf, "*unknown format*");
			break;
	}
}

void storeFields(edb, dict, msg, item, keysonly)
	Empire_Database *edb;
	Dictionary *dict;
	char *msg;
	struct genobject *item;
	int keysonly;
{
	Field *fld;
	int i, subscript, arraysize;
	s_char *bp;
	s_char *cp;
	s_char c;
	char buf[128];

	/* store fields one by one */
	for (fld=dict->flds;fld;fld=fld->next) {

	    /* each array element is handled separately */
	    arraysize=fld->fldcount;
	    if (fld->fldtype==EDB_CHAR) arraysize=0;
#ifdef TAGFIELDS
		/* a minor optimization would be possible if we know key fields.. */
		/* this requires that we know whether a field is key */
		if (keysonly&&!fld->iskey) {
			for (;;) {
				c = *msg++;
				if (c==0||c==':'||c=='\n') break;
			}
			if (*msg==':') msg++;
			continue;
		}
#endif
	    for (subscript = 0; !subscript || subscript<arraysize; ++subscript) {
		/* extract field by copying to buf */
		bp = buf;
		for (i=0;i<127;i++) {
			c = *msg;
			if (c==0||c==':'||c=='\n') break;
			if (arraysize && c==',') break;
			*bp++ = c;
			msg++;
		}
		*bp++ = '\0';
		if (bp!=(s_char *)buf)
			storeField(dict->objdesc, item, fld, buf, dict->objnum, subscript);
		if (arraysize && *msg==',') msg++;
	    }
	    if (*msg==':') msg++;
	}
	
}

struct genobject *newGenObject(edb, desc)
	Empire_Database *edb;
	ObjectDescriptor *desc;
{
	struct genobject *item;
	/* allocate memory for a new object */
	item = (struct genobject *) malloc(desc->objsize);
	if (item)
		memset(item, 0, desc->objsize);	/* initialize it */
	return item;
}

void freeGenObject(edb, dict, item)
	Empire_Database *edb;
	Dictionary *dict;
	struct genobject *item;
{
	/*
	   go through the object and free any attached strings first.
	   then we free the object itself.
	*/
	Field *fld;
	int fldoffs;
	long fldkind;
	Thing thing;
	struct castr *captr;
	
	for (fld=dict->flds;fld;fld=fld->next) {
		/* get object descriptor */	
		captr = fld->boundto;
		if (!captr) continue;
		fldoffs = (captr->ca_code)&0xffff;
		if (fldoffs==0) continue;
		fldkind = (captr->ca_code)&NSC_TMASK;
		if (fldkind!=NSC_CHARP) continue;
		thing.cp = ((char *)item) + fldoffs;
		if (*thing.ccp)
			free(*thing.ccp);
		*thing.ccp = 0;
	}
	free(item);
}

struct genobject *storeObject(edb, code, msg)
	Empire_Database *edb;
	int code;
	char *msg;
{
	Dictionary *dict;
	ObjectDescriptor *desc;
	struct genobject *item, *orig_item;
	int len;
	
	dict = findDict(edb, code);
	if (!dict) {
		/* we never received a definition */
		DebugPrint(("csync: unknown csync type: %d\n", code));
		return 0;
	}
	desc = dict->objdesc;

	/* if we cannot store these kind of objects, just ignore */
	if (!desc) {
		DebugPrint(("csync: no physical representation for object: %d\n", code));
		return 0;
	}

	item = newGenObject(edb, desc);
	item->ef_type = code;

	/* pack the key fields into the object */
	storeFields(edb, dict, msg, item, SF_KEYFIELDSONLY);

	/* see if the object is already in the database */
	orig_item = desc->find(desc, edb, item);

	if (orig_item) {
		/* already there... Just update it with new values */
		if (edb->objbuffer) {
			/* save old info */
			len = edb->objbuflen;
			if (len>desc->objsize) len = desc->objsize;
			memcpy(edb->objbuffer, orig_item, len);
		}
		storeFields(edb, dict, msg, orig_item, SF_ALLFIELDS);
		freeGenObject(edb, dict, item);
		item = orig_item;
	}
	else {
		/* it's a whole new object */
		/* now store the item in the database... */
#ifdef TAGFIELDS
		storeFields(edb, dict, msg, item, SF_ALLFIELDS);
#endif
		desc->store(desc, edb, item);
	}
	return item;
}

/* edb_parse_c_sync -- process a Csync message and update the database */
struct genobject *edb_parse_c_sync(edb, msg)
	Empire_Database *edb;
	s_char *msg;
{
	int code;
	assert(edb!=0);
	assert(msg!=0);
	
	/* preclear a bit of the old object buffer */
	if (edb->objbuffer)
	    if (edb->objbuflen>=4)
		memset(edb->objbuffer, 0, 4);

	/* skip over protocol character and blank if not already done */
	if (strncmp(msg,"f ",2)==0) msg += 2; 
	
	code = getnum(&msg);
	if (code==EDB_TYPE_DEF) {
		parse_typedef(edb, msg+1);
		return 0;
	}
	return storeObject(edb, code, msg+1);
}

void dumpDict(edb)
	Empire_Database *edb;
{
	Dictionary *dict;
	Field *fld;
	for (dict=(Dictionary *)edb->dictionary; dict; dict=dict->next) {
		printf("csync: object dictionary for type %d\n", dict->objnum);
		for (fld=dict->flds;fld;fld=fld->next) {
			printf("	%10s: %d\n", fld->fldname, fld->fldtype);
		}
	}
}

void dumpObject(edb, thing)
	Empire_Database *edb;
	struct genobject *thing;
{
	Dictionary *dict;
	Field *fld;
	char buf[80];
	if (!thing) {
		printf("dumpObject: NULL object\n");
		return;
	}
	dict = findDict(edb, thing->ef_type);
	if (!dict) {
		printf("dumpObject: unknown object type: %d\n", thing->ef_type);
		return;
	}
	printf("	object type = %d\n", thing->ef_type);
	for (fld=dict->flds;fld;fld=fld->next) {
		sprintField(dict->objdesc, buf, thing, fld, thing->ef_type, 0);
		printf("	%10s: %s\n", fld->fldname, buf);
	}
}


